{-# LANGUAGE DeriveDataTypeable, FlexibleInstances, MultiParamTypeClasses, TypeFamilies #-}
module Happstack.Facebook.Friends where

import Control.Applicative
import Control.Applicative.Error
import Data.Traversable (sequenceA)
import Data.Generics (Data, Typeable)
import Data.Ix (Ix)
import qualified Data.Map as Map
import Data.Maybe (catMaybes)
import Happstack.Facebook.Common
import Text.RJson

-- * Friends

-- ** Friends.areFriends


data AreFriends
    = AreFriends  [(User, User)]
      deriving (Eq, Ord, Read, Show, Data, Typeable)

instance (HasSessionKey m) => FacebookMethod m AreFriends where
    type FacebookResponse AreFriends = [(User, User, Maybe Bool)]
    parseResponse _ jstr = parseAreFriends jstr
    toParams (AreFriends friends) =
        do let (users1, users2) = unzip friends
           sessionKey <- askSessionKey
           return $ catMaybes $ 
                      [ Just ("method", "Friends.areFriends")
                      , Just ("session_key", sessionKey)
                      , Just ("uids1", show $ JDArray $ map (JDNumber . fromIntegral . uid) users1)
                      , Just ("uids2", show $ JDArray $ map (JDNumber . fromIntegral . uid) users2)
                      ]


tAreFriends = "[{\"uid1\":222332,\"uid2\":222333,\"are_friends\":true},{\"uid1\":1240077,\"uid2\":1240079,\"are_friends\":false}]"
tAreFriends' = "[{\"uid1\":1240077,\"uid2\":1240078,\"are_friends\":null}]"

parseAreFriends :: String -> Either String (FacebookResponse AreFriends)
parseAreFriends jstr =
    case parseJsonString jstr of
      (Left e) -> Left ("Friends.areFriends: " ++ e)
      (Right json) ->
          case json of
            (JDArray objects) ->
                    case sequenceA $ map toAreFriends objects of
                      Success a -> Right a
                      Failure errs -> Left (unlines errs)
            o -> Left ("Friends.areFriends: expected an array" ++ show o)
        where
          toAreFriends (JDObject map) =
              case (Map.lookup "uid1" map, Map.lookup "uid2" map, Map.lookup "are_friends" map) of
                (Just (JDNumber uid1D), Just (JDNumber uid2D), Just (JDBool areFriends)) ->
                      Success (User $ round uid1D, User $ round uid2D, Just areFriends)
                (Just (JDNumber uid1D), Just (JDNumber uid2D), Just JDNull) ->
                      Success (User $ round uid1D, User $ round uid2D, Nothing)
                _ -> Failure ["Could not parse: " ++ show map]


-- ** Friends.get

-- |see alse GetNS
data Get
    = Get (Maybe Flid)
      deriving (Eq, Ord, Read, Show, Data, Typeable)

instance (HasSessionKey m) => FacebookMethod m Get where
    type FacebookResponse Get = [User]
    parseResponse _ jstr = 
        case reads jstr of
          [(friends,[])] -> Right $ map User friends
          _ -> Left $ "Friends.get: could not parse list: " ++ jstr
    toParams (Get mFlid) =
        do sessionKey <- askSessionKey
           return $ catMaybes $ 
                      [ Just ("method", "Friends.get")
                      , Just ("session_key", sessionKey)
                      , fmap (\(Flid flid) -> ("flid", show flid)) mFlid
                      ]

-- |This version does not require a session key
data GetNS
    = GetNS User (Maybe Flid)
      deriving (Eq, Ord, Read, Show, Data, Typeable)

instance (Functor m, Monad m) => FacebookMethod m GetNS where
    type FacebookResponse GetNS = [User]
    parseResponse _ jstr = 
        case reads jstr of
          [(friends,[])] -> Right $ map User friends
          _ -> Left $ "Friends.get: could not parse list: " ++ jstr
    toParams (GetNS user  mFlid) =
        do return $ catMaybes $ 
                      [ Just ("method", "Friends.get")
                      , Just ("uid", show $ uid user)
                      , fmap (\(Flid flid) -> ("flid", show flid)) mFlid
                      ]

-- ** Friends.getAppUsers

data GetAppUsers
    = GetAppUsers
      deriving (Eq, Ord, Read, Show, Data, Typeable)

instance (HasSessionKey m) => FacebookMethod m GetAppUsers where
    type FacebookResponse GetAppUsers = [User]
    parseResponse _ "{}" = Right []
    parseResponse _ jstr = 
        case reads jstr of
          [(friends,[])] -> Right $ map User friends
          _ -> Left $ "Friends.getAppUsers: could not parse list: " ++ jstr
    toParams GetAppUsers =
        do sessionKey <- askSessionKey
           return $ catMaybes $ 
                      [ Just ("method", "Friends.getAppUsers")
                      , Just ("session_key", sessionKey)
                      ]

-- ** Friends.getLists

data GetLists
    = GetLists
      deriving (Eq, Ord, Read, Show, Data, Typeable)

instance (HasSessionKey m) => FacebookMethod m GetLists where
    type FacebookResponse GetLists = [FriendList]
    parseResponse _ jstr =
        case parseJsonString jstr of
          (Left e) -> Left ("Friends.getLists: " ++ e)
          (Right json) ->
              case json of
                (JDArray objects) -> 
                    case sequenceA $ map toFriendList objects of
                      Success a -> Right a
                      Failure errs -> Left (unlines errs)
                o -> Left ("Friends.getLists: expected an array" ++ show o)
        where
          toFriendList (JDObject map) =
              case (Map.lookup "flid" map, Map.lookup "name" map) of
                (Just (JDNumber flidD), Just (JDString nameStr)) ->
                      Success (FriendList { flid = Flid (round flidD), name = nameStr })
                o -> Failure ["Could not parse as friend list: " ++ show map]

    toParams GetLists =
        do sessionKey <- askSessionKey
           return $ catMaybes $ 
                      [ Just ("method", "Friends.getLists")
                      , Just ("session_key", sessionKey)
                      ]

t =  "[{\"flid\":524434806284,\"name\":\"another sample\"},{\"flid\":524434806283,\"name\":\"sample list\"}]"

-- ** Friends.getMutualFriends

data GetMutualFriends
    = GetMutualFriends User (Maybe User)
      deriving (Eq, Ord, Read, Show, Data, Typeable)

instance (HasSessionKey m) => FacebookMethod m GetMutualFriends where
    type FacebookResponse GetMutualFriends = [User]
    parseResponse _ jstr =
        case reads jstr of
          [(friends,[])] -> Right $ map User friends
          _ -> Left $ "Friends.getMutualFriends: could not parse list: " ++ jstr

    toParams (GetMutualFriends targetUser mSrcUser) =
        do sessionKey <- askSessionKey
           return $ catMaybes $ 
                      [ Just ("method", "Friends.getMutualFriends")
                      , Just ("session_key", sessionKey)
                      , Just ("target_uid", show $ uid targetUser)
                      , fmap (\user -> ("source_uid", show $ uid user)) mSrcUser
                      ]

data GetMutualFriendsNS
    = GetMutualFriendsNS User User
      deriving (Eq, Ord, Read, Show, Data, Typeable)

instance (Functor m, Monad m) => FacebookMethod m GetMutualFriendsNS where
    type FacebookResponse GetMutualFriendsNS = [User]
    parseResponse _ jstr =
        case reads jstr of
          [(friends,[])] -> Right $ map User friends
          _ -> Left $ "Friends.getMutualFriends: could not parse list: " ++ jstr

    toParams (GetMutualFriendsNS targetUser srcUser) =
        do return $ catMaybes $ 
                      [ Just ("method", "Friends.getMutualFriends")
                      , Just ("target_uid", show $ uid targetUser)
                      , Just ("source_uid", show $ uid srcUser)
                      ]
